#ifndef __EXC_C
#error This file must be included from ../exc_1.c
#endif

#ifndef __DJGPP__
#error This file is for djgpp only
#endif

#include <signal.h>
#include <setjmp.h>
#include <crt0.h>
#include <go32.h>
#include <dpmi.h>
#include <stubinfo.h>
#include <sys/resource.h>
#include <sys/farptr.h>
#include <sys/exceptn.h>

#define INT_DIV    0       /* Divide exception */
#define INT_DBG    1       /* Debug exception */
#define INT_NMI    2       /* Non-maskable interrupt */
#define INT_INT3   3       /* Breakpoint */
#define INT_INTO   4       /* Overflow */
#define INT_BND    5       /* Array BOUND interrupt */
#define INT_ILL    6       /* Invalid opcode */
#define INT_DNA    7       /* FPU not present */
#define INT_DBLF   8       /* Double fault */
#define INT_CPSEG  9       /* FPU Segment overrun */
#define INT_ITSS   10      /* Invalid TSS */
#define INT_SNP    11      /* Segment not present */
#define INT_STKF   12      /* Stack fault */
#define INT_PROTF  13      /* General protection */
#define INT_PAGEF  14      /* Page fault */
#define INT_CPE    16      /* FPU error */


static char *excName[] = {
            "Divide Exception", "Debug Exception",       /* 0  */
            "NMI !!??",         "Breakpoint",            /* 2  */
            "Overflow",         "Array Bound Check",     /* 4  */
            "Invalid Opcode",   "FPU not present",       /* 6  */
            "Double Fault",     "FPU Segment Overrun",   /* 8  */
            "Invalid TSS",      "Segment not present",   /* 10 */
            "Stack Fault",      "General Protection",    /* 12 */
            "Page Fault",       "Reserved !!??",         /* 14 */
            "FPU Error",        "Alignment Check",       /* 16 */
          };

static SigFunc oldSigHandler [NUM_EX];

/*
 *  Application interface to exception handler system.
 *  Called once during main start. Add handlers for SIGSEGV, SIGFPE,
 *  SIGTRAP and SIGNOFP. Can be invoked multiple times without hurting.
 */
void InstallExcHandler (excHook hook)
{
  static int been_here = 0;

  if (been_here)
     return;

  been_here   = 1;
  excAppStart = FindAppStart (_my_cs());

  if (excStkEnd == (ULONG)-1)  /* stack limits not set by app */
  {
    struct rlimit rlimit;
    getrlimit (RLIMIT_STACK,&rlimit);
    excStkEnd = rlimit.rlim_cur;
  }

  if (excAppEnd == 0xFFFFFFFF)
     GetCodeLimit (&excAppEnd);

#if STACKREWIND
  InitStackTrace (GetArg0 (_stubinfo->psp_selector));
#endif

  excHookUser = hook;
  videoMode   = _farpeekb (_dos_ds,0x449);
  videoHeight = _farpeekb (_dos_ds,0x484);  /* assume EGA+ */

  oldSigHandler[0] = signal (SIGSEGV, (void(*)(int))FaultingFinish);
  oldSigHandler[1] = signal (SIGFPE,  (void(*)(int))FaultingFinish);
  oldSigHandler[2] = signal (SIGTRAP, (void(*)(int))FaultingFinish);
  oldSigHandler[3] = signal (SIGNOFP, (void(*)(int))FaultingFinish);
  oldSigHandler[4] = signal (SIGILL,  (void(*)(int))FaultingFinish);
  atexit (RemoveExcHandler);
}

/*
 *  Fill out exception register structure from exception jmp_buf
 */
static excReg *ExcFillContext (jmp_buf ctx)
{
  static excReg reg;
  reg.DS  = ctx[0].__ds;
  reg.ES  = ctx[0].__es;
  reg.FS  = ctx[0].__fs;
  reg.GS  = ctx[0].__gs;
  reg.SS  = ctx[0].__ss;
  reg.CS  = ctx[0].__cs;
  reg.EIP = ctx[0].__eip;
  reg.EAX = ctx[0].__eax;
  reg.EBX = ctx[0].__ebx;
  reg.ECX = ctx[0].__ecx;
  reg.EDX = ctx[0].__edx;
  reg.ESI = ctx[0].__esi;
  reg.EDI = ctx[0].__edi;
  reg.EFL = ctx[0].__eflags;
  reg.ESP = ctx[0].__esp;
  reg.EBP = ctx[0].__ebp;
  reg.CR0 = 0;
  reg.CR2 = 0;
  reg.excNumber = ctx[0].__signum;
  reg.excCode   = ctx[0].__sigmask & 0xFFFF;
  return (&reg);
}

void RemoveExcHandler (void)
{
  static int been_here = 0;

  if (!been_here)
  {
    been_here = 1;
    signal (SIGSEGV,oldSigHandler[0]);
    signal (SIGFPE, oldSigHandler[1]);
    signal (SIGTRAP,oldSigHandler[2]);
    signal (SIGNOFP,oldSigHandler[3]);
    signal (SIGILL, oldSigHandler[4]);
  }
}

/*
 *  Since there's no standard way for getting at "char *argv[]"
 *  passed to "main()", we must do it ourself
 */
char *GetArg0 (USHORT env_sel)
{
#if 0
  /* for now we trust `__crt0_argv' is correct
   * later, we should peek at DOS's cmd-line.
   */
#else
  return (__crt0_argv[0]);
#endif
}

/*
 * Find the first allocated page. I.e find first valid address of CS.
 */
static UINT FindAppStart (USHORT sel)
{
  return (4096);  /* to-do: query DPMI host; __dpmi_get_page_attributes() */
}

/*
 * Return to text-mode if video-mode changed since we
 * installed ourself. Also check for 50-line mode.
 */
static void SetTextMode (void)
{
  if (_farpeekb(_dos_ds,0x449) != videoMode)
  {
    __dpmi_regs reg;
    reg.x.ax = 0x83;
    __dpmi_int (0x10,&reg);
    if (videoHeight > 25)        /* was 50-line screen     */
    {
      reg.x.ax = 0x1112;         /* load 8x8 font, block 0 */
      reg.x.bx = 0;
      __dpmi_int (0x10,&reg);
    }
  }
}

/*
 * Return segment limits of CS and DS selectors
 */
void GetCodeLimit (ULONG *limit)
{
  *limit = __dpmi_get_segment_limit (_my_cs());
}

void GetDataLimit (ULONG *limit)
{
  *limit = __dpmi_get_segment_limit (_my_ds());
}

#if AUTO_INIT
/*
 * Define this function as constructor. This will be called
 * before main() is called.
 */
void ctor_ExcHandler (void)
{
  InstallExcHandler (NO_EXC_HOOK);
}

__asm__(".section .ctor; .long _ctor_ExcHandler; .section .text");

#endif

