
TITLE Exception handler for PharLap 386|DosX
NAME  EXC_0

INCLUDE exc.inc

PUBLIC @_context,       _context
PUBLIC @_oldExcHandlers,_oldExcHandlers
PUBLIC @_ExcFillContext,_ExcFillContext

_DATA SEGMENT

  PUB_DW excCseg, SS_CODE+3
  PUB_DW excDseg, SS_DATA+3

  PUB_DD excHookMain, 0                 ; address of main exc-handler

  PUB_DD exc_BOS, ?
         dd 1023 dup (?)                ; exception handler stack

  PUB_DD exc_TOS, 0                     ; stack start

      exc_removed     dd 0
      exit_stack      dd 0,0            ; address of stack to use for exit()
      exit_func       dd 0,0            ; address of main exit() func

     _context         label dword
    @_context         ExcStruc <>

     _oldExcHandlers  label dword       ; MASM 5.0 doesn't grok `PWORD'
    @_oldExcHandlers  dp NUM_EX dup (0) ; FARPTR oldExcHandlers[18];

_DATA ENDS


_TEXT SEGMENT

;
; Pharlap's High-level interface for user installed exception handlers:
;
; Stack layout:   ofs  38h   CR2       At time of orignal int
;                      34h   INT #     Int/Exception number
;                      30h   DosxFlag  For DosX umbrella handler
;                      2Ch   GS        At time of orignal int
;                      28h   FS        At time of orignal int
;                      24h   DS        At time of orignal int
;                      20h   ES        At time of orignal int
;                      1Ch   SS        At time of orignal int
;                      18h   ESP       At time of orignal int
;                      14h   EFLAGS    At time of orignal int
;                      10h   CS        At time of orignal int
;                      0Ch   EIP       At time of orignal int
;                      08h   EFLAGS    For DosX umbrella handler
;                      04h   CS        For DosX umbrella handler
;           SS:ESP->   00h   EIP       For DosX umbrella handler
;           SS:ESP->  -04h   Code      Exception code for exc 8,10-14

_Rmode      equ 1                         ; int occurred in real mode
_LDSregs    equ 2                         ; load seg regs on return

;
; Stack layout referred to ss:[ebx]. EBX = entry ESP.
;
ex_CR2      equ dword ptr ss:[ebx+38h]  ; CR2 at time of interrupt
ex_INUM     equ dword ptr ss:[ebx+34h]  ; interrupt number
ex_FL       equ dword ptr ss:[ebx+30h]  ; flag bits, value 0-3
ex_GS       equ dword ptr ss:[ebx+2Ch]  ; GS at time of interrupt
ex_FS       equ dword ptr ss:[ebx+28h]  ; FS at time of interrupt
ex_DS       equ dword ptr ss:[ebx+24h]  ; DS at time of interrupt
ex_ES       equ dword ptr ss:[ebx+20h]  ; ES at time of interrupt
ex_SS       equ dword ptr ss:[ebx+1Ch]  ; SS:ESP at time of interrupt
ex_ESP      equ dword ptr ss:[ebx+18h]  ;
ex_EFLAGS   equ dword ptr ss:[ebx+14h]  ; EFLAGS at time of interrupt
ex_CS       equ dword ptr ss:[ebx+10h]  ; CS:EIP at time of interrupt
ex_EIP      equ dword ptr ss:[ebx+0Ch]  ;
ex_DOSX_EFL equ dword ptr ss:[ebx+08h]  ; return stack frame in
ex_DOSX_CS  equ dword ptr ss:[ebx+04h]  ; 386|DOS-Extender handler
ex_DOSX_EIP equ dword ptr ss:[ebx+00h]  ; return address
ex_ICODE    equ dword ptr ss:[ebx-04h]  ; exception error code

;
; NOTE: The exception handler is currently for flat-model only
;

;---------------------------------------------------------------

C_PROC  ExcGlue_0                      ; ISR for exception 0-7,9,15-17
        push eax                       ; save registers used
        push ebx
        mov  ebx, esp
        add  ebx, 8                    ; EBX -> DOSX_EIP

        mov  ax, cs     ; !!
        add  ax, 8      ; !!           ; DS = CS+8
        mov  ds, ax     ; !!

        mov  _context.excCode, 0       ; no exception code
        call @_ExcFillContext          ; fill ExcStruc
        call @patchReturn              ; patch return address
        pop  ebx
        pop  eax
        iretd                          ; IRET and continue at @excCont below

C_PROC  ExcGlue_1                      ; ISR for exception 8, 10-14
        push eax                       ; save registers used
        push ebx
        mov  ebx, esp
        add  ebx, 12                   ; EBX -> DOSX_EIP

        mov  ax, cs     ; !!
        add  ax, 8      ; !!           ; DS = CS+8
        mov  ds, ax     ; !!

        mov  eax, ex_ICODE             ; save exception code
        mov  _context.excCode, eax
        call @_ExcFillContext          ; fill ExcStruc
        call @patchReturn              ; patch return address
        pop  ebx
        pop  eax
        add  esp, 4                    ; discard error code
        iretd                          ; IRET and continue at @excCont

;-------------------------------------------------------------------
;
;  DosX umbrella handler tranfers control to this point
;

@excCont:
        cmp _excHookMain, 0            ; HLL hook installed ?
        je   @toDosx                   ; no, back to Dosx
        sti
        lea  esp, @_exc_TOS            ; setup a safe stack
        call [_excHookMain]            ; yes, do main exception handling
      ;
      ; We normally never gets here. If HLL-hook deinstalled
      ; exc-vectors, we'll get a new exception from the DosX here
      ;
        mov esp, _context._ESP         ; get old ESP, EBP unchanged
        jmp [_context._EIP]            ; return to faulty EIP

@toDosx:push eax
        push ecx
        mov  eax, _context.excNum      ; get exception number
        mov  cl, al
        imul eax, 6                    ; far pointer is 6 bytes
        lea  eax, _oldExcHandlers[eax]
        push ds
        mov  edx, [eax]
        mov  ds, [eax+4]               ; DS:EDX = oldExcHandler [excNum]
        mov  eax, 2533h                ; set exception vector
        int  21h
        pop  ds
        mov  eax, _context._EIP        ; get EIP for exception
        mov  [esp+8], eax              ; put on stack
        pop  ecx
        pop  eax
        ret                            ; return to cause of exception
                                       ; and let DosX handle it

;-------------------------------------------------------------------

ex_EBX equ dword ptr [esp+4]           ; EBX saved on stack
ex_EAX equ dword ptr [esp+8]           ; EAX saved on stack

C_PROC  ExcFillContext                 ; also called by sig_0()

        mov  eax, ex_INUM              ; load exception number
        mov  _context.excNum, eax      ; store it
        mov  _context._ECX, ecx        ; store other registers
        mov  _context._EDX, edx
        mov  _context._EDI, edi
        mov  _context._ESI, esi
        mov  _context._EBP, ebp

        mov  eax, ex_CR2
        mov  _context._CR2, eax        ; CR2 register
        mov  eax, ex_EAX
        mov  _context._EAX, eax        ; EAX register
        mov  eax, ex_EBX
        mov  _context._EBX, eax        ; EBX register
        mov  eax, ex_EIP
        mov  _context._EIP, eax        ; EIP register
        mov  eax, ex_CS
        mov  _context._CS, ax          ; CS register
        mov  eax, ex_EFLAGS
     ;  and  eax, not 10000h           ; clear resume flag
        or   eax, 10000h               ; set resume flag
        mov  _context._EFlags, eax     ; EFlag register
        mov  eax, ex_ESP
        mov  _context._ESP, eax        ; ESP register
        mov  eax, ex_SS
        mov  _context._SS, ax          ; SS register
        mov  eax, ex_ES
        mov  _context._ES, ax          ; ES register
        mov  eax, ex_DS
        mov  _context._DS, ax          ; DS register
        mov  eax, ex_FS
        mov  _context._FS, ax          ; FS register
        mov  eax, ex_GS
        mov  _context._GS, ax          ; GS register
        ret

@patchReturn:
        mov  eax, offset cs:@excCont
        mov  ex_EIP, eax               ; patch EIP resume location
        and  ex_FL, not _LDSregs       ; don't reload segment registers
     ;  mov  ax, cs
     ;  add  ax, 8
     ;  mov  ds, ax                    ; reload DS/ES incase <> SS_DATA
     ;  mov  es, ax
        ret

;------------------------------------------------------------------
;
; void exc_write (int handle, char *buf, int len);
;
C_PROC  exc_write
        push ebx
        mov  ebx, [esp+8]     ; handle
        mov  edx, [esp+12]    ; buf
        mov  ecx, [esp+16]    ; len
        mov  eax, 4000h       ; DOS write to handle
        int  21h
        pop  ebx
        ret

;
; void exc_close (int handle);
;
C_PROC  exc_close
        push ebx
        mov  ebx, [esp+8]     ; handle
        mov  eax, 3E00h       ; DOS close handle
        int  21h
        pop  ebx
        ret

;
; void cdecl init_exit (void (*fn)(), UINT stk);
;
C_PROC  init_exit
        mov  eax, [esp+4]
        mov  dword ptr exit_func[0], eax
        mov  ax, cs
        mov  word ptr exit_func[4], ax

        mov  eax, [esp+8]
        mov  dword ptr exit_stack[0], eax
        mov  ax, ds
        mov  word ptr exit_stack[4], ax
        ret

;
; void cdecl exc_exit (int status);
;
; Does a far call to C Run-time exit handler
;
C_PROC  exc_exit
        mov  ecx, [esp+4]
        lss  esp, pword ptr exit_stack
        mov  ax, ss
        mov  ds, ax
        mov  es, ax           ; Flat model, i.e. ES=DS=SS

        push ecx
        push ecx
        jmp  pword ptr exit_func

;
; void cdecl raw_exit (int status);
;
; Exits straight to DOS
;
;C_PROC  raw_exit
;        pop  eax          ; discard return address
;        pop  eax
;        mov  ah, 4Ch
;        int  21h

;
; void cdecl RemoveExcHandler (void);
;
C_PROC  RemoveExcHandler
        xor  ecx, ecx
        cmp  exc_removed, 0
        je   @next
        ret
        
@next:  mov  eax,ecx
        imul eax, 6                    ; far pointer is 6 bytes
        lea  eax, _oldExcHandlers[eax]
        push ds
        mov  edx, [eax]
        mov  ds, [eax+4]
        mov  eax, 2533h                ; set exception vector
        int  21h
        pop  ds
        inc  ecx
        cmp  ecx, NUM_EX
        jb   @next

        mov  exc_removed, 1
        ret
;
; int cdecl exc_alloc (ULONG pages, USHORT *sel);
;
; Allocate a selector with specified # of pages

C_PROC  exc_alloc
        push ebx
        mov  ebx, [esp+8]     ; EBX = # of pages
        mov  ah, 48h
        int  21h
        jc   @alloc_fail      
        mov  ebx, [esp+12]    ; store allocated selector
        mov  [ebx], ax
        mov  eax, 1           ; ret 1 (success)
        pop  ebx
        ret
@alloc_fail:
        xor  eax, eax         ; ret 0 (fail)
        pop  ebx
        ret

_TEXT ENDS
END

