/*
 * Simple POSIX signal handling for Pharlap
 *
 */
#include <stdlib.h>
#include <setjmp.h>
#include <signal.h>
#include <process.h>

#define __EXC_INTERNAL
#include "exc.h"
#include "printk.h"
#include "ioport.h"
#include "peek.h"

#if !defined(__MSC32__)

#define STATIC
#define SIGMAX 33

STATIC void   SigInt_8rm (void);
STATIC void   SigInt_9rm (void);
STATIC void   SigInt_1B  (void);
STATIC void   SigInt_1C  (void);
STATIC void   SigInt_23  (void);

STATIC void   LocalSignalHook (void);
STATIC void   Cleanup_SIGINT  (void);
STATIC void   Cleanup_SIGALRM (void);

static void   (*oldHook)(void) = NULL;
static volatile char gotSignal = 0;
static int      sig_timer      = 0;
static ULONG    sig_countdown  = 0;
static ULONG    sig_reload     = 0;

static RMC_BLK  rmRegs;
static SigFunc  signalList [SIGMAX];       /* SIG_DFL = 0 */

#if defined(__HIGHC__)
#  define SIG_LO SIGHUP
#  define SIG_HI SIGUSR3
   static char signames[] =
     "HUP   INT   QUIT  ILL   TRAP  ABRT  EMT   FPE   KILL  BUS   "\
     "SEGV  SYS   PIPE  ALRM  TERM  URG   STOP  TSTP  CONT  CHLD  "\
     "TTIN  TTOU  IO    XCPU  XFSZ  VTALRMPROF  WINCH USR1  USR2  USR3";

#elif defined(__BORLANDC__)
#  define SIG_LO  SIGINT
#  define SIG_HI  SIGABRT
#  define SIGTRAP 5
#  define SIGALRM 14
   static char signames[] =
     "INT   undef ILL   TRAP  undef undef FPE   undef undef "\
     "SEGV  undef undef ALRM  TERM  USR1  USR2  undef undef "\
     "USR3  BREAK ABRT  ";

#elif defined(__WATCOMC__)
#  define SIG_LO  SIGABRT
#  define SIG_HI  SIGBREAK
#  define SIGTRAP 8
#  define SIGALRM 9
   static char signames[] =
      "ABRT  FPE   ILL   INT   SEGV  TERM  BREAK" \
      "TRAP  ALRM";

#elif defined(__MSC32__)
#  define SIG_LO  SIGABRT
#  define SIG_HI  SIGBREAK
#  define SIGTRAP 8
#  define SIGALRM 9
   static char signames[] =
      "ABRT  FPE   ILL   INT   SEGV  TERM  BREAK" \
      "TRAP  ALRM";

#else
#  error Define SIG names for your compiler
#endif

#define ARGSUSED(x) (x = x)

static int excToSignal [NUM_EX] = {   /* Hardware signals */
             SIGILL, SIGTRAP,SIGILL, SIGTRAP,SIGILL,SIGILL,
             SIGSEGV,SIGFPE, SIGILL, SIGFPE, SIGILL,SIGSEGV,
             SIGSEGV,SIGSEGV,SIGSEGV,SIGILL, SIGFPE,SIGSEGV
           };

PMRM_INTR oldintr_8, oldintr_9, oldintr_1B, oldintr_1C, oldintr_23;

/*----------------------------------------------------------------------*/

SigFunc _signal (int sig, SigFunc func)
{
  SigFunc temp;

  if (sig <= 0 || sig > SIGMAX)
     return (SIG_ERR);

  InstallExcHandler (NO_EXC_HOOK);
  if (oldHook == NULL)
      oldHook = excHookMain;
  excHookMain = LocalSignalHook;

  if (sig == SIGINT)
  {
    _dx_install_intr ((rmodeIntr)SigInt_9rm, SigInt_9pm, 9, &oldintr_9);
    _dx_install_intr ((rmodeIntr)SigInt_1B, NULL, 0x1B, &oldintr_1B);
    _dx_install_intr ((rmodeIntr)SigInt_23, NULL, 0x23, &oldintr_23);
    atexit (Cleanup_SIGINT);
  }
#if defined(__HIGHC__)
  else if (sig == SIGALRM || sig == SIGUSR1)
  {
    _dx_install_intr ((rmodeIntr)SigInt_8rm, SigInt_8pm, 8, &oldintr_8);
//  _dx_install_intr ((rmodeIntr)SigInt_1C, NULL, 0x1C, &oldintr_1C);
    atexit (Cleanup_SIGALRM);
  }
#else
  else if (sig == SIGUSR1)
  {
    _dx_install_intr ((rmodeIntr)SigInt_8rm, SigInt_8pm, 8, &oldintr_8);
    atexit (Cleanup_SIGALRM);
  }
#endif

  temp = signalList [sig-1];
  signalList [sig-1] = func;
  return (temp);
}

/*----------------------------------------------------------------------*/

int _raise (int sig)
{
  SigFunc handler;

  if (sig <= 0 || sig > SIGMAX)
     return (-1);

  if ((handler = signalList[sig-1]) == (SigFunc)SIG_IGN)
     return (0);                      /* ignore it */

  if (handler != (SigFunc)SIG_DFL &&
      ((ULONG)handler > excStkStart || (ULONG)handler < excAppStart))
  {
    fdprintk (2, "Bad signal handler %08X\r\n", (void*)handler);
    handler = (SigFunc)SIG_DFL;
  }

  if (handler == (SigFunc)SIG_DFL)    /* no user handler, use default */
  {
    if (sig >= SIG_LO && sig <= SIG_HI)
         fdprintk (2, "raising signal SIG%.5s\r\n", signames+6*(sig-SIG_LO));
    else fdprintk (2, "raising signal %d\r\n.", sig);

    sig = excToSignal [context.excNumber];
    if (sig == SIGSEGV || sig == SIGTRAP || sig == SIGFPE)
       (*oldHook)();
    exit (1);
  }
  handler (sig);    /* call user's signal handler */
  return (0);
}

/*----------------------------------------------------------------------*/

int _kill (int pid, int sig)
{
  if (pid == getpid() || pid == 0 ||
      pid == -getpid() || pid == -1)
  {
    if (sig)
       _raise (sig);
    return (0);
  }
  return (-1);
}

/*----------------------------------------------------------------------*/

int _pause (void)
{
  while (!gotSignal)
        ;

  gotSignal = 0;
  return (0);
}

/*----------------------------------------------------------------------*/

unsigned _alarm (unsigned sec)
{
  if (sec == 0)
  {
    sig_timer = 0;
    return (0);
  }
  sig_countdown = (ULONG) (18.2 * (double)sec);
  sig_reload = sig_countdown;
  sig_timer  = 1;
  return (sec);
}

/*----------------------------------------------------------------------*/

STATIC void LocalSignalHook (void)
{
  static  int reentered = -1;
  static  excReg *r = &context;
  jmp_buf jb;
  int     sig;

  if (++reentered)
  {
    fdprintk (2,
              "PANIC: nested exception (%d at %02X:%08X) in signal handling\r\n",
              r->CS, r->EIP, r->excNumber);
    exit (-1);
  }
  gotSignal = 1;
  sig = excToSignal [r->excNumber];

  if (_raise(sig) == -1)
     (*oldHook)();

#if defined (__BORLANDC__)
  jb[0].j_ebp = r->EBP;
  jb[0].j_ebx = r->EBX;
  jb[0].j_edi = r->EDI;
  jb[0].j_esi = r->ESI;
  jb[0].j_esp = r->ESP;
  jb[0].j_ret = r->EIP;
#elif defined (__HIGHC__)
  jb[0]  = r->EIP;
  jb[1]  = r->CS;
  jb[2]  = r->EBP;
  jb[3]  = r->ESP;
  jb[4]  = r->DS;
  jb[5]  = r->ES;
  jb[6]  = r->FS;
  jb[7]  = r->GS;
  jb[8]  = r->ESI;
  jb[9]  = r->EDI;
  jb[10] = r->EBX;
#elif defined (__WATCOMC__)
  jb[0]  = r->EBX;
  jb[1]  = r->ECX;
  jb[2]  = r->EDX;
  jb[3]  = r->ESI;
  jb[4]  = r->EDI;
  jb[5]  = r->EBP;
  jb[6]  = r->ESP;
  jb[7]  = r->ES;
  jb[8]  = r->DS;
  jb[9]  = r->CS;
  jb[10] = r->FS;
  jb[11] = r->GS;
  jb[12] = r->SS;

#else
#  error I don't know jmp_buf layout for this compiler
#endif
  --reentered;
  longjmp (jb, r->EAX);
}

/*-------------------------------------------------------------------*/

STATIC void WriteFan (void)
{
  static char index = 0;
  PokeRealWord (0xB8000000+158, 0x0E00 + "-\\|/"[index++ & 3]);
}

/*--- Functions for supporting SIGALRM ------------------------------*/

STATIC void Cleanup_SIGALRM (void)
{
  _dx_cleanup_intr (&oldintr_8);
  _dx_cleanup_intr (&oldintr_1C);
}

void dump_context (void)
{
  excReg *r = &context;

  fdprintk (2,"\r\nException %d at %02X:%08X\r\n",
            r->excNumber, r->CS, r->EIP);

  fdprintk (2,
            "Regs: EAX %08X  EBX %08X  ECX %08X  EDX %08X\r\n"
            "      ESI %08X  EDI %08X  EBP %08X  ESP %08X\r\n"
            "      DS  %04X      FS  %04X      SS  %04X\r\n"
            "      ES  %04X      GS  %04X      FLG %08X\r\n",
            r->EAX, r->EBX, r->ECX, r->EDX,
            r->ESI, r->EDI, r->EBP, r->ESP,
            r->DS,  r->FS,  r->SS,
            r->ES,  r->GS,  r->EFL);
}

void SigInt_8 (void)
{
  _dx_call_real (oldintr_8.rmVector,&rmRegs,1);

  if (sig_timer && sig_countdown-- == 0)
  {
    gotSignal = 1;
  //_raise (SIGALRM);
    *(UCHAR*)context.EIP = 0xCC;  /* int 3 */
    sig_countdown = sig_reload;
    dump_context();
  }
}

STATIC void SigInt_8rm (void)
{
  WriteFan();
  SigInt_8();
}

STATIC void SigInt_1C (void)
{
  WriteFan();
  if (sig_timer && sig_countdown-- == 0)
  {
    gotSignal = 1;
    _raise (SIGALRM);
    sig_countdown = sig_reload;
  }
}

/*--- Functions for supporting SIGINT -------------------------------*/

STATIC void Cleanup_SIGINT (void)
{
  _dx_cleanup_intr (&oldintr_9);
  _dx_cleanup_intr (&oldintr_1B);
  _dx_cleanup_intr (&oldintr_23);
}

STATIC void KeyboardAck (void)
{
  volatile UCHAR k = _inportb (0x61);
  _outportb (0x61, k | 0x80);                /* toggle keyboard ..    */
  _outportb (0x61, k);                       /* .. acknowledge line   */
  _outportb (0x20, 0x20);                    /* Send end-of-interrupt */
  pokew (0, 0x41C, peekw(0,0x41A));          /* flush keyboard buffer */
}

void SigInt_9 (void)
{
  UCHAR ctrl = peekb (0,0x417) & 15;
  UCHAR scan = _inportb (0x60);

  _dx_call_real (oldintr_9.rmVector, &rmRegs, 1);

  if ((ctrl & 4) && (scan == 0x2E))  /* ^C pressed */
  {
    KeyboardAck();
    gotSignal = 1;
  //_raise (SIGINT);
    *(UCHAR*)context.EIP = 0xCC;  /* int 3 */
    dump_context();
  }
}

STATIC void SigInt_9rm (void)
{
  WriteFan();
  SigInt_9();
}

STATIC void SigInt_1B (void)
{
  gotSignal = 1;
  context.excNumber = 20;
  _raise (SIGINT);
}

STATIC void SigInt_23 (void)
{
  gotSignal = 1;
  context.excNumber = 19;
  _raise (SIGINT);
}

#endif  /* __MSC32__ */
